/*============================================================================
 * make_books.C: makes an index file (stdout) and a set of book review web
 * pages, descending the directory tree looking for *.rev files.
 *----------------------------------------------------------------------------
 * Copyright 2001 Tom Hull.
 *
 * This is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *---------------------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define ERR strerror(errno)

class book {
public:
    book(const char *p);
   ~book();

    FILE *open();

    //  flag values
    enum { norev = 0x01 };

    char *path;
    char *html;
    char *title;
    char *author;
    char *asort;
    char *year;
    char *pub;
    time_t stime;
    time_t htime;
    unsigned flag;
    book *next;
};

class bclass {
public:
    bclass(const char *);
   ~bclass();

    char *name;
    book *head;
};

bclass *bclasses[256];
int     nbclass;

FILE *outf = stdout;
FILE *find;

void writeHdr(FILE *f, const char *title);
void writeEnd(FILE *f);
void processFile(const char *path);
void addClass(book *bp, const char *p);
void writeClasses();
int sortBooks(const void *p1, const void *p2);
void lccat(char *b, const char *p, const char *e);

int main(int argc, char **argv)
{
    int c;
    char pbuf[PATH_MAX], *p;

    while ((c = getopt(argc, argv, "")) != -1) {
        switch (c) {
        default:
            printf("usage: make_books\n");
            return 1;
        }
    }

    writeHdr(outf, "Tom Hull: Book Review Index");

    find = popen("find . -name '*.rev' -print", "r");
    while (fgets(pbuf, sizeof pbuf, find)) {
        if (p = strchr(pbuf, '\n'))
            *p = 0;
        processFile(pbuf);
    }
    pclose(find);

    writeClasses();

    writeEnd(outf);

    return 0;
}

void writeHdr(FILE *f, const char *title)
{
    fprintf(f, "<html>\n");
    fprintf(f, "<head>\n");
    fprintf(f, "<title>%s</title>\n", title);
    fprintf(f, "</head>\n\n");

    fprintf(f, "<body>\n");
}

void writeEnd(FILE *f)
{
    fprintf(f, "</body>\n");
    fprintf(f, "</html>\n");
}

void processFile(const char *path)
{
    FILE *f, *outf;
    char buf[1024];
    book *bp = new book(path);

    if (f = bp->open()) {
        if (bp->stime >= bp->htime && !(bp->flag & book::norev)) {
            // need to rewrite .html file
            fprintf(stderr, "building: %s\n", bp->html);
            if ((outf = fopen(bp->html, "w")) == 0) {
                fprintf(stderr, "cannot create: %s [%s]\n", bp->html, ERR);
            }
            else {
                sprintf(buf, "%s: %s", bp->author, bp->title);
                writeHdr(outf, buf);
                fprintf(outf, "<h3><b>%s: <i>%s</i></b></h3>\n",
                        bp->author, bp->title);
                fprintf(outf, "%s, %s<p>\n", bp->pub, bp->year);
                fprintf(outf, "<hr>\n");

                while (fgets(buf, sizeof buf, f)) {
                    fputs(buf, outf);
                }

                writeEnd(outf);
                fclose(outf);
            }
        }
        fclose(f);
    }
}

void addClass(book *bp, const char *p)
{
    int i;
    bclass *cp = 0;
    book **bpp;

    for (i = 0; i < nbclass; i++) {
        cp = bclasses[i];
        if (!strcmp(cp->name, p)) {
            break;
        }
    }
    if (i == nbclass) {
        bclasses[nbclass++] = cp = new bclass(p);
    }
    for (bpp = &cp->head; *bpp; bpp = &(*bpp)->next);
    *bpp = bp;
}

void writeClasses()
{
    int i, j, n;
    bclass *cp;
    book *bbuf[256], *bp;

    for (i = 0; i < nbclass; i++) {
        cp = bclasses[i];
        printf("<h3>%s</h3>\n", cp->name);
        printf("<ul>\n");
        for (n = 0, bp = cp->head; bp; bp = bp->next) {
            bbuf[n++] = bp;
        }
        qsort(bbuf, n, sizeof bbuf[0], sortBooks);
        for (j = 0; j < n; j++) {
            bp = bbuf[j];
            if (bp->flag & book::norev) {
                printf("<li><b>%s: <i>%s</i></b>\n", bp->author, bp->title);
            }
            else {
                printf("<li><a href=\"%s\"><b>%s: <i>%s</i></b></a>\n",
                       bp->html, bp->author, bp->title);
            }
        }
        printf("</ul>\n");
    }
}

int sortBooks(const void *p1, const void *p2)
{
    int n;
    const book *b1 = *((const book **) p1);
    const book *b2 = *((const book **) p2);

    n = strcmp(b1->asort, b2->asort);
    if (n == 0)
        n = strcmp(b1->year, b2->year);
    return n;
}

book::book(const char *p)
{
    char buf[1024], *r;
    struct stat st;

    flag = 0;
    path = strdup(p);
    //  concoct html pathname
    if (p = strrchr(p, '/'))
        p++;
    else
        p = path;
    strcpy(buf, p);
    if ((r = strrchr(buf, '.')) == 0)
        r = buf + strlen(buf);
    strcpy(r, ".html");
    html = strdup(buf);
    //  stat both files
    stime = (stat(path, &st) == 0) ? st.st_mtime : 0;
    htime = (stat(html, &st) == 0) ? st.st_mtime : 0;
    //  null missing fields
    title = author = asort = year = pub = 0;
    next = 0;
}

book::~book()
{
    //  should free buffered fields.
}

FILE *book::open()
{
    FILE *f;
    char buf[1024], *p;

    if ((f = fopen(path, "r")) == 0) {
        fprintf(stderr, "cannot open: %s [%s]\n", path, ERR);
        return 0;
    }
    while (fgets(buf, sizeof buf, f)) {
        if (p = strchr(buf, '\n'))
            *p = 0;
        if (p = strchr(buf, '|')) {
            *p++ = 0;
            switch (buf[0]) {
            case 'A':
                if (!strcmp(buf, "Author"))
                    author = strdup(p);
                else if (!strcmp(buf, "Asort"))
                    asort = strdup(p);
                else
                    goto err;
                break;
            case 'C':
                if (!strcmp(buf, "Class"))
                    addClass(this, p);
                else
                    goto err;
                break;
            case 'N':
                if (!strcmp(buf, "NoRev"))
                    flag |= norev;
                else
                    goto err;
                break;
            case 'P':
                if (!strcmp(buf, "Pub"))
                    pub = strdup(p);
                else
                    goto err;
                break;
            case 'T':
                if (!strcmp(buf, "Title"))
                    title = strdup(p);
                else
                    goto err;
                break;
            case 'Y':
                if (!strcmp(buf, "Year"))
                    year = strdup(p);
                else
                    goto err;
                break;
            default:
            err:
                fprintf(stderr, "bad key: %s [%s]\n", buf, path);
                break;
            }
        }
        else
            break;
    }
    if (author && !asort) {
        buf[0] = 0;
        if (p = strrchr(author, ' ')) {
            lccat(buf, p, 0);
            lccat(buf, author, p);
        }
        else
            lccat(buf, author, 0);
        asort = strdup(buf);
    }
    return f;
}

bclass::bclass(const char *p)
{
    name = strdup(p);
    head = 0;
}

bclass::~bclass()
{
    //  should free up buffered memory
}

void lccat(char *b, const char *p, const char *e)
{
    if (e == 0)
        e = p + strlen(p);
    while (*b)
        b++;
    for ( ; p < e; p++) {
        if (isalpha(*p))
            *b++ = tolower(*p);
    }
    *b = 0;
}

